"""
An Executor instance executes actions immediately based on scheduler names
supplied by time- or threshold-based schedulers. The actions associated with
the named scheduler are executed.
"""

from typing import Callable
import logging
from .exception import TerminateSchedulerException

logger = logging.getLogger(__name__)


class Executor:
    def __init__(
        self, get_actions_thunk: Callable, unschedule_scheduler_thunk: Callable
    ):
        self.get_actions_thunk = get_actions_thunk
        self.unschedule_scheduler_thunk = unschedule_scheduler_thunk

    def push(self, scheduler_name: str):
        """
        Event triggering (time- or threshold-based) calls this method.
        """
        actions_dictionary = self.get_actions_thunk(scheduler_name)
        for action_name in actions_dictionary:
            tag = f"{scheduler_name}:{action_name}"
            action = actions_dictionary[action_name]
            try:
                action.execute(tag=tag)
                logger.info(f"Execution: tag ({tag}); executed action ({action})")
            except TerminateSchedulerException as terminate:
                self.unschedule_scheduler_thunk(scheduler_name)
                logger.info(
                    f"Execution: tag ({tag}); unscheduled scheduler ({scheduler_name}); TerminateSchedulerException raised ({str(terminate)})"
                )
            except Exception as exception:
                logger.exception(
                    f"Execution: tag ({tag}); error while executing action ({action})",
                    exc_info=exception,
                )



from fastapi import APIRouter, status, Depends
from whendo.core.util import DateTime2
from whendo.api.shared import return_success, raised_exception, get_dispatcher
from whendo.core.resolver import resolve_program

router = APIRouter(prefix="/programs", tags=["Programs"])


@router.get("/deferred_program_count", status_code=status.HTTP_200_OK)
def get_expiring_action_count():
    try:
        return return_success(
            {
                "deferred_program_count": get_dispatcher(
                    router
                ).get_deferred_program_count()
            }
        )
    except Exception as e:
        raise raised_exception(f"failed to retrieve the deferred program count", e)


@router.get("/clear_deferred_programs", status_code=status.HTTP_200_OK)
def clear_deferred_actions():
    try:
        get_dispatcher(router).clear_all_deferred_programs()
        return return_success("deferred programs were cleared")
    except Exception as e:
        raise raised_exception("failed to clear deferred programs", e)


@router.get("/{program_name}", status_code=status.HTTP_200_OK)
def get_program(program_name: str):
    try:
        program = get_dispatcher(router).get_program(program_name=program_name)
        return return_success(program)
    except Exception as e:
        raise raised_exception(f"failed to retrieve the program ({program_name})", e)


@router.post("/{program_name}", status_code=status.HTTP_200_OK)
def add_program(program_name: str, program=Depends(resolve_program)):
    try:
        assert program, f"couldn't resolve class for program ({program_name})"
        get_dispatcher(router).add_program(program_name=program_name, program=program)
        return return_success(f"program ({program_name}) was successfully added")
    except Exception as e:
        raise raised_exception(f"failed to add program ({program_name})", e)


@router.put("/{program_name}", status_code=status.HTTP_200_OK)
def set_program(program_name: str, program=Depends(resolve_program)):
    try:
        assert program, f"couldn't resolve class for program ({program_name})"
        get_dispatcher(router).set_program(program_name=program_name, program=program)
        return return_success(f"program ({program_name}) was successfully updated")
    except Exception as e:
        raise raised_exception(f"failed to update program ({program_name})", e)


@router.delete("/{program_name}", status_code=status.HTTP_200_OK)
def delete_program(program_name: str):
    try:
        get_dispatcher(router).delete_program(program_name=program_name)
        return return_success(f"program ({program_name}) was successfully deleted")
    except Exception as e:
        raise raised_exception(f"failed to delete program ({program_name})", e)


@router.get("/{program_name}/describe", status_code=status.HTTP_200_OK)
def describe_program(program_name: str):
    try:
        return get_dispatcher(router).describe_program(program_name=program_name)
    except Exception as e:
        raise raised_exception(f"failed to describe program ({program_name})", e)


@router.post("/{program_name}/schedule", status_code=status.HTTP_200_OK)
def schedule_program(program_name: str, datetime2: DateTime2):
    try:
        dt1 = datetime2.dt1
        dt2 = datetime2.dt2
        get_dispatcher(router).schedule_program(
            program_name=program_name, start=dt1, stop=dt2
        )
        return return_success(f"program ({program_name}) was successfully scheduled")
    except Exception as e:
        raise raised_exception(f"failed to schedule program ({program_name})", e)


@router.get("/{program_name}/unschedule", status_code=status.HTTP_200_OK)
def unschedule_program(program_name: str):
    try:
        get_dispatcher(router).unschedule_program(program_name=program_name)
        return return_success(f"program ({program_name}) was successfully unscheduled")
    except Exception as e:
        raise raised_exception(f"failed to schedule program ({program_name})", e)



from typing import Dict, Any, Optional
import os
import logging
from typing import Optional
from whendo.core.util import PP, Dirs, Now, SystemInfo
from whendo.core.action import Action

logger = logging.getLogger(__name__)


class FileAppend(Action):
    """
    This action appends supplied stuff to a file. The mode field
    determines what to do with the field, payload, and the argument,
    data.

    P -- use payload (P). if absent use data (D).
    D -- use D. if absent use P.
    PD -- use P containing D if both. P if just P. D if just D.
    DP -- use D containing P if both. D if just D. P if just P.

    Note: any mix of lower and upper case are valid modes.

    Usage: two ways
        1. FileAppend using <mode> field
        2. one of FileAppendP, FileAppendD, FileAppendPD, FileAppendDP
    """

    file_append: str = "file_append"
    file: str
    payload: Optional[dict]
    relative_to_output_dir: bool = True
    mode: str = "PD"

    def description(self):
        return f"This action appends <payload> and/or <data> to ({self.file})."

    def execute(self, tag: str = None, data: dict = None):
        payload = self.compute_payload(self.mode, payload=self.payload, data=data)
        file = (
            os.path.join(Dirs.output_dir(), self.file)
            if self.relative_to_output_dir
            else self.file
        )
        with open(file, "a") as outfile:
            PP.pprint(payload, stream=outfile)
            outfile.write("\n")
        # return {"result": payload, "action_info": self.info()}
        return self.action_result(
            result=f"file ({self.file}) appended", data=data, extra=self.info()
        )

    def compute_payload(
        self, mode: str, payload: Optional[dict] = None, data: dict = None
    ):
        result = {}
        upper_mode = mode.upper()
        if upper_mode == "P":
            if payload:
                result = payload
            elif data:
                result = data
        elif upper_mode == "D":
            if data:
                result = data
            elif payload:
                result = payload
        elif upper_mode == "DP":
            if data:
                result = data
                if payload:
                    result["payload"] = payload
            elif payload:
                result = payload
        elif upper_mode == "PD":
            if payload:
                result = payload
                if data:
                    result["data"] = data
            elif data:
                result = data
        else:
            raise ValueError(
                f"invalid file append mode ({self.mode}); should be one of 'P', 'D', 'PD', 'DP"
            )
        return result


class FileAppendP(FileAppend):
    """ one of the two, with payload having precedence"""

    file_append_p: str = "file_append_p"
    mode = "P"

    def description(self):
        return f"This action appends payload, otherwise data to ({self.file})."


class FileAppendD(FileAppend):
    """ one of the two, with data having precedence"""

    file_append_d: str = "file_append_d"
    mode = "D"

    def description(self):
        return f"This action appends data, otherwise payload to ({self.file})."


class FileAppendPD(FileAppend):
    """ one or both, with data inside payload """

    file_append_pd: str = "file_append_pd"
    mode = "PD"

    def description(self):
        return f"This action appends data inside payload to ({self.file})."


class FileAppendDP(FileAppend):
    """ one or both, with payload inside data """

    file_append_dp: str = "file_append_dp"
    mode = "DP"

    def description(self):
        return f"This action appends payload inside data to ({self.file})."



from typing import List, Dict
import logging
from datetime import datetime, timedelta
from typing import Optional
from whendo.core.program import Program, ProgramItem


logger = logging.getLogger(__name__)


class PBEProgram(Program):
    """
    A prologue/body/epilogue Program has a prologue, body, and
    epilogue.

    1. prologue action name -- invoked at the start time
    2. epilogue action name -- invoked at the stop time
    3. body -- a schedulers/actions-structured dictionary
        These schedulers/actions are deferred until...
            start datetime + offset seconds
        and expire
            stop datetime - offset seconds

    usage:

        pbe = PBEProgram()
            .body_element("heartbeat", "turn_on_pin_A")
            .body_element("heartbeat2", "report_activity")
            .epilogue("clear_gpio")
            .prologue("start_pivot")
        pbe = PBEProgram(
            prologue="start_pivot",
            epilogue="clear_gpio",
            body={"heartbeat": ["turn_on_pin_A"],
                "heartbeat2", ["report_activity"]}
            )
    """

    prologue_name: Optional[str] = None
    epilogue_name: Optional[str] = None
    body: Dict[str, List[str]] = {}
    offset_seconds: int = 0

    def description(self):
        return f"This program starts with action ({self.prologue_name}) and ends with ({self.epilogue_name}) with scheduled actions ({self.body}) in between."

    def compute_program_items(self, start: datetime = None, stop: datetime = None):
        """
        Note: None is supplied when ensuring existence of action and scheduler names in the dispatcher.
        """
        result = []
        if self.prologue_name:
            result.append(
                ProgramItem("defer", start, "immediately", self.prologue_name)
            )

        if len(self.body) > 0:
            timedelta_offset = timedelta(seconds=self.offset_seconds)
            start_plus = start + timedelta_offset if start else None
            stop_minus = stop - timedelta_offset if stop else None
            for scheduler_name in self.body:
                for action_name in self.body[scheduler_name]:
                    result.append(
                        ProgramItem("defer", start_plus, scheduler_name, action_name)
                    )
                    result.append(
                        ProgramItem("expire", stop_minus, scheduler_name, action_name)
                    )

        if self.epilogue_name:
            result.append(ProgramItem("defer", stop, "immediately", self.epilogue_name))

        return result

    def body_element(self, scheduler_name: str, action_name: str):
        if scheduler_name not in self.body:
            self.body[scheduler_name] = [action_name]
        elif action_name not in self.body[scheduler_name]:
            self.body[scheduler_name].append(action_name)
        return self

    def prologue(self, prologue_name: str):
        self.prologue_name = prologue_name
        return self

    def epilogue(self, epilogue_name: str):
        self.epilogue_name = epilogue_name
        return self



from typing import Optional
import logging
from whendo.core.util import TimeUnit
from whendo.core.scheduler import TimedScheduler
from whendo.core.timed import Timed
from whendo.core.executor import Executor

logger = logging.getLogger(__name__)


class Timely(TimedScheduler):
    """
    This scheduler executes Actions every [interval] days, hours, minutes, seconds at a
    time within the time unit (hour, minute, second).

    The time unit is determined by the values of the hour, minute, and second
    fields.

    If hour is specified, the time unit is set to day. [17:00:00]
    If minute is specified, the time unit is set to hour. [:15:00]
    If second is specified, the time unit is set to minute. [:15]
    Otherwise the time unit is set to second.

    For example, for hour=None, minute=15, second=None, interval=2, the system will
    execute the action every 2 hours at 15 minutes past the hour.
    """

    interval: int
    hour: Optional[int] = None
    minute: Optional[int] = None
    second: Optional[int] = None

    def description(self):
        return f"This scheduler executes its actions every {self.time_as_str()}. {super().description()}"

    def schedule(self, scheduler_name: str, executor: Executor):
        self.unschedule(scheduler_name)
        callable = self.during_period(scheduler_name=scheduler_name, executor=executor)
        self.get_timed().schedule_timely_callable(
            tag=scheduler_name,
            callable=callable,
            interval=self.interval,
            hour=self.hour,
            minute=self.minute,
            second=self.second,
        )

    def time_as_str(self):
        hour = self.hour
        minute = self.minute
        second = self.second
        interval = self.interval
        if hour:
            if minute is None:
                minute = 0
            if second is None:
                second = 0
            return f"{interval} day{'s' if interval > 1 else ''} at {hour:02}:{minute:02}:{second:02}."
        elif minute:
            if second is None:
                second = 0
            return f"{interval} hours{'s' if interval > 1 else ''} at :{minute:02}:{second:02} past the hour."
        elif second:
            return f"{interval} minute{'s' if interval > 1 else ''} at :{second:02} past the minute."
        else:
            return f"{interval} second{'s' if interval > 1 else ''}"


class Randomly(TimedScheduler):
    """
    This scheduler randomly executes Actions every [low] to [high] days, hours, minutes, or seconds.

    The time unit is supplied at instance creation.

    For example, for low=30, hight=90 and time_unit=second, the system will again execute the action
    randomly every 30 to 90 seconds after each execution.
    """

    time_unit: TimeUnit = TimeUnit.second
    low: int
    high: int

    def description(self):
        return f"This scheduler executes its actions every {self.low} to {self.high} {self.time_unit.value}s. {super().description()}"

    def schedule(self, scheduler_name: str, executor: Executor):
        self.unschedule(scheduler_name)
        callable = self.during_period(scheduler_name=scheduler_name, executor=executor)
        self.get_timed().schedule_random_callable(
            tag=scheduler_name,
            callable=callable,
            time_unit=self.time_unit,
            low=self.low,
            high=self.high,
        )



"""
This script establishes the top FastAPI path and all sub-path routers.
It also initializes the Dispatcher for the calling ASGI server.

This version is only used in unit testing -- the Dispatcher construction
is different from main.py.
"""
from fastapi import FastAPI
from whendo.core.dispatcher import Dispatcher
from whendo.core.timed import Timed
from whendo.api.router import actions, schedulers, dispatcher, jobs, execution, programs
from whendo.api.shared import set_dispatcher

app = FastAPI()


@app.get("/")
async def root():
    return "whengo API server started (unit test)"


dispatcher_instance = Dispatcher()
dispatcher_instance.set_timed(Timed())
dispatcher_instance.initialize()

app.include_router(set_dispatcher(actions.router, dispatcher_instance))
app.include_router(set_dispatcher(schedulers.router, dispatcher_instance))
app.include_router(set_dispatcher(dispatcher.router, dispatcher_instance))
app.include_router(set_dispatcher(jobs.router, dispatcher_instance))
app.include_router(set_dispatcher(execution.router, dispatcher_instance))
app.include_router(set_dispatcher(programs.router, dispatcher_instance))



"""
This file contains the psutil functions necessary for the absence of
psutil in a whendo environment.

Not having psutil in the build spec might enable

    >> python -m build

to work. The build failed from the presence of psutil as a required
library.
"""

from socket import AddressFamily
from collections import namedtuple
import logging


logger = logging.getLogger(__name__)


def net_if_addrs():
    named_tuple = namedtuple(
        "snicaddr", ["family", "address", "netmask", "broadcast", "ptp"]
    )
    snicaddr = named_tuple(
        AddressFamily.AF_INET, "0.0.0.0", "255.0.0.0", "127.0.0.1", None
    )
    return {"unknown": [snicaddr]}


def virtual_memory():
    svmem = namedtuple("svmem", ["unknown"])
    return svmem(0.0)


def getloadavg():
    return [0.0, 0.0, 0.0]


def cpu_percent():
    return 0.0



import time
import logging
from whendo.core.action import Action
import whendo.core.util as util


logger = logging.getLogger(__name__)


class SysInfo(Action):
    """
    Return system info
    """

    sys_info = "sys_info"

    def description(self):
        return f"This action returns system-level information."

    def execute(self, tag: str = None, data: dict = None):
        return self.action_result(result=util.SystemInfo.get(), data=data)


class MiniInfo(Action):
    """
    Return terse local info.
    """

    mini_info = "mini_info"

    def description(self):
        return f"This action returns terse local information."

    def execute(self, tag: str = None, data: dict = None):
        return self.action_result(result=self.local_info(), data=data)


class Pause(Action):
    """
    Sleep for supplied int seconds. It's best to make sure that the sleep duration
    does not interfere with job execution (blocking the thread).
    """

    pause: str = "pause"
    seconds: float = 1.0

    def description(self):
        return f"This action sleeps for {self.seconds} second{'s' if self.seconds != 1 else ''}."

    def execute(self, tag: str = None, data: dict = None):
        time.sleep(self.seconds)
        return self.action_result(result=self.seconds, data=data)



from typing import List, Dict, Any, Optional
from enum import Enum
import json
import logging
from collections import namedtuple
from whendo.core.action import Action
from whendo.core.exception import TerminateSchedulerException


logger = logging.getLogger(__name__)


def processing_results(
    result: Any,
    processing_count: int = 0,
    success_count: int = 0,
    exception_count: int = 0,
    successful_actions: List[Dict[Any, Any]] = [],
    exception_actions: List[Dict[Any, Any]] = [],
):
    return {
        "result": result,
        "processing_count": processing_count,
        "success_count": success_count,
        "exception_count": exception_count,
        "successful_actions": successful_actions,
        "exception_actions": exception_actions,
    }


class Arg(Action):
    arg: str = "arg"
    data: dict

    def description(self):
        return f"This action returns ({self.data})."

    def execute(self, tag: str = None, data: dict = None):
        return self.data


class Terminate(Action):
    terminate: str = "terminate"

    def description(self):
        return f"This action raises a TerminateScheduler exception."

    def execute(self, tag: str = None, data: dict = None):
        raise TerminateSchedulerException(value=tag)


class Failure(Action):
    """ acts like False """

    failure: str = "failure"

    def description(self):
        return f"This action always fails. It serves a role similar to the bool, False."

    def execute(self, tag: str = None, data: dict = None):
        raise Exception("purposely unsuccessful execution", self.json())


class Success(Action):
    """ acts like True. It also serves as a fixed point execute implementation. """

    success: str = "success"

    def description(self):
        return (
            f"This action always succeeds. It serves a role similar to the bool, True."
        )

    def execute(self, tag: str = None, data: dict = None):
        return data


class Not(Action):
    """ acts like negation """

    _not: str = "not"
    operand: Action

    def description(self):
        return f"If the operand ({self.operand}) fails, Not succeeds, otherwise fails (throwing an Exception). It serves a role similar to logical negation."

    def execute(self, tag: str = None, data: dict = None):
        operand_result = None
        try:
            operand_result = self.operand.execute(tag=tag, data=data)
        except Exception as exception:
            operand_result = exception
        if isinstance(operand_result, Exception):
            extra = {
                "exception": str(operand_result),
                "action": self.info(),
            }
            return self.action_result(result=True, data=data, extra=extra)
        else:
            raise Exception(
                "exception generated; action execution treated as a failure",
                self.json(),
            )


class ListOpMode(str, Enum):
    ALL = "all"
    OR = "or"
    AND = "and"


class ListAction(Action):
    """
    executes actions based on:

        ListOpMode.ALL:   executes all, irrespective of individual outcomes
        ListOpMode.OR:    executes until the first successful action (no exception)
        ListOpMode.AND:   executes until the first exception (failure)

    Intended to be abstract class; not intended to be instantiated. Its subclasses,
    All, Or, and And, should be used instead.
    """

    op_mode: ListOpMode
    actions: List[Action] = []
    exception_on_no_success: bool = False
    result_only: bool = True

    def execute(self, tag: str = None, data: dict = None):
        processing_info = process_actions(
            data=data,
            op_mode=self.op_mode,
            tag=tag,
            actions=self.actions,
            successful_actions=[],
            exception_actions=[],
        )
        if (
            processing_info["processing_count"] > 0
            and processing_info["success_count"] == 0
            and self.exception_on_no_success
        ):
            raise Exception(
                "exception generated; all actions failiing treated as a list action failure",
                json.dumps(self.dict()),
                json.dumps(processing_info),
            )
        return processing_info["result"] if self.result_only else processing_info


class All(ListAction):
    """
    executes all actions
    """

    _all: str = "all"
    op_mode: ListOpMode = ListOpMode.ALL

    def description(self):
        return f"This action executes all of these actions in order: ({self.actions})."

    def execute(self, tag: str = None, data: dict = None):
        return super().execute(tag=tag, data=data)


class Or(ListAction):
    """
    executes actions until first success
    """

    _or: str = "or"
    op_mode: ListOpMode = ListOpMode.OR

    def description(self):
        return f"This action executes all of these actions in order until the first success: ({self.actions}). It serves a role similar to logical or."

    def execute(self, tag: str = None, data: dict = None):
        return super().execute(tag=tag, data=data)


class And(ListAction):
    """
    executes actions until first failure
    """

    _and: str = "and"
    op_mode: ListOpMode = ListOpMode.AND

    def description(self):
        return f"This action executes all of these actions in order until the first failure: ({self.actions}). It serves a role similar to logical and."

    def execute(self, tag: str = None, data: dict = None):
        return super().execute(tag=tag, data=data)


def process_actions(
    op_mode: ListOpMode,
    actions: List[Action],
    tag: str = None,
    data: Dict[Any, Any] = None,
    successful_actions: List[Dict[Any, Any]] = [],
    exception_actions: List[Dict[Any, Any]] = [],
    processing_count: int = 0,
    success_count: int = 0,
    exception_count: int = 0,
):
    """
    services the three action list classes, ALL, OR, and AND
    """
    exceptions: List[Exception] = []
    loop_data = data
    for action in actions:
        try:
            result = action.execute(tag=tag, data=loop_data)
            logger.info(
                f"Execution: tag ({tag}); executed action ({action}); loop data ({loop_data})"
            )
        except Exception as exception:
            result = exception
            logger.exception(
                f"Execution: tag ({tag}); error while executing action ({action}); loop data ({loop_data})",
                exc_info=exception,
            )
        processing_count += 1
        if isinstance(result, Exception):
            exception_count += 1
            exception_dict = action.dict().copy()
            exception_dict.update({"exception": str(result)})
            exception_actions.append(exception_dict)
            if isinstance(result, TerminateSchedulerException):
                result.value = processing_results(
                    loop_data,
                    processing_count,
                    success_count,
                    exception_count,
                    successful_actions,
                    exception_actions,
                )
                raise result
            if op_mode == ListOpMode.AND:  # stop after first failure
                break
            loop_data = exception_dict
        else:
            success_count += 1
            successful_actions.append(action.dict())
            loop_data = result
            if op_mode == ListOpMode.OR:  # stop after first success
                break
    return processing_results(
        loop_data,
        processing_count,
        success_count,
        exception_count,
        successful_actions,
        exception_actions,
    )


# experimental


class Compose(Action):
    compose: str = "compose"
    actions: List[Action]

    def description(self):
        return (
            f"Performs function composition on supplied actions in left-to-right order."
        )

    def execute(self, tag: str = None, data: dict = None):
        loop_data = data
        for action in self.actions:
            loop_data = action.execute(tag=tag, data=loop_data)
        return loop_data
